// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//==============================================================================
// File Name: WeightVector.h
// modifed by Tianbao Yang (01.07.13)
// based on Shai Shalev-Shwartz's file in Pegasos code http://www.cs.huji.ac.il/~shais/code/
// modifications: add customized data, constructors, operators and serializations 
//==============================================================================

#ifndef _WeightVector_H
#define _WeightVector_H

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <sstream>
#include <cmath>
#include <algorithm>    // std::transform
#include <boost/mpi.hpp>
#include "simple_sparse_vec_hash.h"
#include "dense_vec.h"

class WeightVector {
  
 private:
 //WeightVector(const WeightVector&);// disallowed
  
 public:

 // a copy constructor needed storing the class object in vector array
 WeightVector(const WeightVector& W){
   d=W.d;
   my_v =new double[d];
   my_a = W.my_a;
   my_snorm=W.my_snorm;
   my_1norm=W.my_1norm;
   for(uint i=0; i<d; ++i)
	     my_v[i]=W.my_v[i];
 }

 //an  assignment operator,  commented out the original assignment operator
  WeightVector& operator=( const WeightVector& W ) {
	  // change for boost MPI
	  if(d!=W.d){
		  delete [] my_v;
		  d=W.d;
		  my_v =new double[d];
	  }
	  if(this != &W){
		  my_a=W.my_a;
		  my_snorm=W.my_snorm;
		  my_1norm=W.my_1norm;
		  for( size_t i = 0; i < d; ++i )
			  my_v[ i ] = W.my_v[ i ];
	  }
	  return *this;
	  
	  /*if (d != W.d) {
	      std::cerr << "Assigning WeightVector of size "
			<< W.d << " to a WeightVector of size "
			<< d << " is not permitted" << std::endl;
	      exit(EXIT_FAILURE);
	    }
	    else if( this != &W ) {
	          my_a=W.my_a;
	          my_snorm=W.my_snorm;
	          my_1norm=W.my_1norm;
	          for( size_t i = 0; i < d; ++i )
	              my_v[ i ] = W.my_v[ i ];
	      }
	    return *this;*/
   }



 // Construct a Weight vector with dimension d
  WeightVector(uint dd) : d(dd), my_a(1.0), my_snorm(0.0), my_1norm(0.0), my_v(NULL){
    my_v = new double[d];
    for (uint i=0; i<d; ++i) my_v[i] = 0.0;
  }

 //default construct with no input, adding for boost MPI
  WeightVector(): d(0), my_a(0.0), my_snorm(0.0), my_1norm(0.0), my_v(NULL){
  }
  

  // Construct a Weight vector with dimension d from a file
  WeightVector(uint dd,std::ifstream &model_file) : d(dd), my_a(1.0), my_snorm(0.0), my_1norm(0.0), my_v(NULL) {
    my_v = new double[d];
    for (uint i=0; i<d; ++i) my_v[i] = 0.0;
    unsigned int ind = 0;
    model_file >> ind;
    while (model_file.good()) {
      char c;
      model_file >> c;
      double val;
        if(ind>=d){
            std::cerr<<"initial model error!"<<std::endl;
            	  exit(EXIT_FAILURE);}
           
      model_file >> val;
      my_v[ind] = val;
      my_snorm += val*val;
      my_1norm +=fabs(val);
      model_file >> ind;
    }
  }

  WeightVector(std::ifstream &model_file) : d(0), my_a(1.0), my_snorm(0.0), my_1norm(0.0), my_v(NULL) {
	model_file>>d;  
    my_v = new double[d];
    for (uint i=0; i<d; ++i) my_v[i] = 0.0;
    unsigned int ind = 0;
    model_file >> ind;
    while (model_file.good()) {
      char c;
      model_file >> c;
      double val;
        if(ind>=d){
            std::cerr<<"initial model error!"<<std::endl;
            	  exit(EXIT_FAILURE);}
           
      model_file >> val;
      my_v[ind] = val;
      my_snorm += val*val;
      my_1norm +=fabs(val);
      model_file >> ind;
    }
  }

  WeightVector(std::istringstream &model_file) : d(0), my_a(1.0), my_snorm(0.0), my_1norm(0.0), my_v(NULL) {
	model_file>>d;  
    my_v = new double[d];
    for (uint i=0; i<d; ++i) my_v[i] = 0.0;
    unsigned int ind = 0;
    model_file >> ind;
    while (model_file.good()) {
      char c;
      model_file >> c;
      double val;
        if(ind>=d){
            std::cerr<<"initial model error!"<<std::endl;
            	  exit(EXIT_FAILURE);}
           
      model_file >> val;
      my_v[ind] = val;
      my_snorm += val*val;
      my_1norm +=fabs(val);
      model_file >> ind;
    }
  }
  
  // destructor
  ~WeightVector() { delete[] my_v;  }

  // this *= s
  void scale(double s) { 
    my_snorm *= (s*s); 
    my_1norm *= fabs(s);
    if (s != 0.0) {
      my_a *= s; 
    } else {
      my_a = 1.0;
      for (uint i=0; i<d; ++i) my_v[i] = 0.0;
    }
  }
  
  //overload plus operator for boost MPI
  WeightVector operator+(const WeightVector& W ) const {
	if (this->d != W.d) {
	  	      std::cerr << "Adding WeightVector of size "
	  			<< W.d << " to a WeightVector of size "
	  			<< d << " is not permitted" << std::endl;
	  	      exit(EXIT_FAILURE);
	}
	WeightVector sW(this->d);
	double swsnorm=0.0;
	double sw1norm=0.0; 
    for(uint i=0; i<d; i++) {
       sW.my_v[i]=this->my_a*this->my_v[i] + W[i];
       swsnorm +=sW.my_v[i]*sW.my_v[i];
       sw1norm +=fabs(sW.my_v[i]);
    }
    sW.my_a=1.0;
    sW.my_snorm=swsnorm;
    sW.my_1norm=sw1norm;
    return sW;
  }
  
  // this += s*x
  void add(simple_sparse_vector& x, double s);

  // this += s*x
  void add(dense_vector& x, double s);
  
  // this += s*x
  void add(WeightVector& x, double s);
  
  template<typename T>
  void add(std::vector<T>& x, double s);

  
  // this -=s*x
  void minus(simple_sparse_vector&x, double s);
  
  // this -=s*x
  void minus(WeightVector&x, double s);

  // this -=s*x
  void minus(dense_vector&x, double s);
  
  template<typename T>
  void minus(std::vector<T>& x, double s);


  

  // w =\min_u 1/2|u - w | + s |u|_1
  double proximal_l1(double s);

  void print(std::ostream& os);

  double operator[](uint i) const{
    if (i < d) return (my_v[i] * my_a);
    return 0.0;
  }

  uint dimension() {return d;}

  // ||this||_1
  double l1norm(){
	  return my_1norm;
  }

  // ||this||^2
  double snorm() {
    return my_snorm;
  }

  double a() {
    return my_a;
  }

  // copy other
 /* void operator= (const WeightVector& other) {
    if (d != other.d) {
      std::cerr << "Assigning WeightVector of size "
		<< other.d << " to a WeightVector of size "
		<< d << " is not permitted" << std::endl;
      exit(EXIT_FAILURE);
    }
    my_a = other.my_a; my_snorm = other.my_snorm;
    for (uint i=0; i<d; ++i) my_v[i] = other.my_v[i];
  }*/

// copy from an array
  void copy(double* u, double scale) {
	my_a=1.0;
	my_snorm=0.0;
	my_1norm=0.0;
    for (uint i=0; i<d; ++i)
    	{
    	   my_v[i] = scale*u[i];
    	   my_snorm+=my_v[i]*my_v[i];
    	   my_1norm+=fabs(my_v[i]);
    	}
  }

  // copy from an array
    void Setsnorm() {
  	my_snorm=0.0;
  	my_1norm=0.0;
      for (uint i=0; i<d; ++i)
      	{
      	   my_snorm+=my_v[i]*my_v[i];
      	   my_1norm+=fabs(my_v[i]);
      	}
      my_snorm*=my_a*my_a;
      my_1norm*=fabs(my_a);
    }


  // make_my_a_one
  // use it for forcing my_a to be 1.0
  double* make_my_a_one() {
    for (uint i=0; i<d; ++i) {
      my_v[i] *= my_a;
    }
    my_a = 1.0;
    return my_v;
  }


 private:  
  // The internal representation of w is as w = a*v where:
  uint d;
  double my_a; 
  double my_snorm;
  double my_1norm;
  double* my_v;
  
  
  // add serialization for boost MPI
  friend class boost::serialization::access; 
  template<class Archive> 
  void save(Archive & ar, const unsigned int version) const { 
       ar & d & my_a & my_snorm & my_1norm;
       for (uint i = 0; i <d; ++i) 
         ar & my_v[i]; 
  } 
  template<class Archive> 
  void load(Archive & ar, const unsigned int version) { 
     ar & d & my_a & my_snorm & my_1norm; 
     assert(d >=0); 
     if (d > 0) { 
       my_v = new double[d]; 
       for (uint i = 0; i < d; ++i) 
           ar & my_v[i]; 
     } 
   } 
   BOOST_SERIALIZATION_SPLIT_MEMBER();
  
  
};


struct vecW_plus
 { // functor for operator+ 
    std::vector<WeightVector> operator()(const std::vector<WeightVector>& _Left, const std::vector<WeightVector>& _Right) const 
    { // apply operator+ to operands 
        std::vector<WeightVector> v(_Left.size()); 
        std::transform(_Left.begin(), _Left.end(), _Right.begin(), v.begin(), std::plus<WeightVector>()); 
        return (v); 
    } 
 };


//-----------------------------------------------------------------------------
/** Operator * for vector-vector multiplication
    @param u A reference to a simple_sparse_vector
    @param v A reference to a WeightVector
    @return The product (double)
*/
double operator* (simple_sparse_vector& u, WeightVector& v);
double operator* (dense_vector& u, WeightVector& v);
template<typename T>
double operator* (std::vector<T>& u, WeightVector& v);

//-----------------------------------------------------------------------------
/** Operator * for vector-vector multiplication
    @param v A reference to a WeightVector
    @param u A reference to a simple_sparse_vector
    @return The product (double)
*/
double operator* (WeightVector& v, simple_sparse_vector& u);
double operator* (WeightVector& v, dense_vector& u);
template<typename T>
double operator* (WeightVector& v, std::vector<T>& u);


double add_return_snorm(WeightVector &u, WeightVector& v);

double add_return_snorm(WeightVector &u, WeightVector& v, double s);

double add_return_snorm(WeightVector &u, WeightVector& v, double s1, double s2);

long get_runtime(void);

#endif
